@@ -1,3 +1,4 @@
+import inspect
 import abc
 from collections import defaultdict, deque
 from uuid import uuid4
@@ -9,10 +10,13 @@ class Effect(metaclass=abc.ABCMeta):
     """
     Base class for effects which can be yielded to the tapystry event loop.
     """
-    def __init__(self, type, oncancel=(lambda: None), name=None):
+    def __init__(self, type, oncancel=(lambda: None), name=None, caller=None):
         self.type = type
         self.cancel = oncancel
         self.name = name
+        if caller is None:
+            caller = inspect.stack()[2]
+        self._caller = caller
 
     def __str__(self):
         if self.name is not None:
@@ -134,7 +138,7 @@ _noval = object()
 
 
 class Strand():
-    def __init__(self, gen, args=(), kwargs=None, parent=None):
+    def __init__(self, caller, gen, args=(), kwargs=None, parent=None):
         if kwargs is None:
             kwargs = dict()
         self._it = gen(*args, **kwargs)
@@ -149,6 +153,13 @@ class Strand():
             self._result = self._it
             self._done = True
         self._effect = None
+        if self._parent is None:
+            self._parent_effect = None
+        else:
+            self._parent_effect = self._parent[0]._effect
+            assert self._parent_effect is not None
+
+        self._caller = caller
 
     def send(self, value=None):
         assert not self._canceled
@@ -162,6 +173,20 @@ class Strand():
             self._result = e.value
             self._effect = None
             return dict(done=True)
+        except Exception as e:
+            tb = e.__traceback__.tb_next
+            line = tb.tb_lineno
+            # line = tb.tb_frame.f_code.co_firstlineno
+            # line number is not exactly right?
+            raise TapystryError(
+                "\n".join([
+                    f"Exception caught at",
+                    f"{self.stack()}",
+                    f":",
+                    f"File {tb.tb_frame.f_code.co_filename}, line {line}, in {tb.tb_frame.f_code.co_name}",
+                    f"{type(e).__name__}: {e}",
+                ])
+            )
 
     def __hash__(self):
         return self.id.int
@@ -170,12 +195,32 @@ class Strand():
         return f"Strand[{self.id.hex}] (waiting for {self._effect})"
 
     def stack(self):
+        # if self._parent is None:
+        #     return [f"Strand[{self.id.hex}]"]
+        # else:
+        #     stack = list(self._parent[0].stack())
+        #     stack.append(f"{self._parent[1]} Strand[{self.id.hex}]")
+        #     return stack
+
+        s = "\n".join([
+            f"File {self._caller[1]}, line {self._caller[2]}, in {self._caller[3]}",
+            f"  {self._caller[4][0].strip()}",
+        ])
         if self._parent is None:
-            return [f"Strand[{self.id.hex}]"]
+            return s
         else:
-            stack = list(self._parent[0].stack())
-            stack.append(f"{self._parent[1]} Strand[{self.id.hex}]")
-            return stack
+            return "\n".join([
+                self._parent[0].stack(),
+                f"Yields effect {self._parent_effect}, created at",
+                s
+            ])
+
+        # print(self._stack[2][0])
+        # print(self._stack[2][1])  # file
+        # print(self._stack[2][2])  # line
+        # print(self._stack[2][3])  # name
+        # print(self._stack[2][4])  # code
+
 
     def is_done(self):
         return self._done
@@ -224,7 +269,7 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False):
     # list of intercept items
     intercepts = []
 
-    initial_strand = Strand(gen, args, kwargs)
+    initial_strand = Strand(inspect.stack()[1], gen, args, kwargs)
     if initial_strand.is_done():
         # wasn't even a generator
         return initial_strand.get_result()
@@ -363,7 +408,7 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False):
         elif isinstance(effect, Receive):
             add_waiting_strand("broadcast." + effect.key, item.strand, effect.predicate)
         elif isinstance(effect, Call):
-            strand = Strand(effect.gen, effect.args, effect.kwargs, parent=(item.strand, effect.name or "call"))
+            strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=(item.strand, effect.name or "call"))
             item.strand._children.append(strand)
             if strand.is_done():
                 # wasn't even a generator
@@ -372,7 +417,7 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False):
                 add_waiting_strand("done." + strand.id.hex, item.strand)
                 advance_strand(strand)
         elif isinstance(effect, CallFork):
-            fork_strand = Strand(effect.gen, effect.args, effect.kwargs, parent=(item.strand, effect.name or "fork"))
+            fork_strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=(item.strand, effect.name or "fork"))
             item.strand._children.append(fork_strand)
             advance_strand(item.strand, fork_strand)
             if not fork_strand.is_done():
@@ -386,7 +431,7 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False):
         elif isinstance(effect, Sleep):
             advance_strand(item.strand)
         else:
-            raise TapystryError(f"Unhandled effect type {type(effect)}")
+            raise TapystryError(f"Unhandled effect type {type(effect)}: {item.strand.stack()}")
 
     for strand in hanging_strands:
         if not strand.is_canceled():
